/*
 * $Project: dedit Disk EDITor for X680x0 $
 * $Copyright: 1989-1993 by d(T.Kawamoto) $
 * $Source: /home/src/cvs/dedit/nonstd.c,v $
 * $Author: kawamoto $
 * $Revision: 1.2 $
 * $Date: 1993/10/24 18:37:29 $
 * $Log: nonstd.c,v $
 *	Revision 1.2  1993/10/24  18:37:29  kawamoto
 *	*** empty log message ***
 *
 *	Revision 1.1  1993/10/24  17:42:48  kawamoto
 *	version 2.24 first cvs importing
 */

#include "dedit.h"
#include "console.h"
#include "diskbios.h"
#include "nonstd.h"
#include "debug.h"

unsigned short timer_max = 1000;

#define UNFORMATED	0
#define SINGLE_DENS	1
#define DOUBLE_DENS	2
#define MIXED_DENS	3

#define TRACK_MAX	160 /* 154 */

typedef struct IDBUFFER {
  uchar C, H, R, N;
  uchar density, duplicate;
  ushort counter;
} idbuf;

static idbuf dblBuf[80], sglBuf[40], buf[120];
static uchar density;
static int dblMaxNum, sglMaxNum, wrkMaxNum, maxNum, offset;

static int sub1(idbuf *);
static void sub2(void);
static CMD_PARAM_BUF cmd;

int NonStdReadID()
{
  int work1, work2;

  cmd.cmd = CMD_SEEK;
  DPRINT("seek ",0);
  diskbios(&cmd);
  DPRINT("end ",0);
  cmd.cmd = CMD_READ_ID;
  DPRINT("read ID ",0);
  if (diskbios(&cmd) & return_ICL)
    {
      DPRINT("end ",0);
      cmd.cmd = CMD_READ_ID | OPT_MFM_MODE;
      DPRINT("read ID MFM mode ",0);
      if (diskbios(&cmd) & return_ICL)
	{
	  DPRINT("end ",0);
	  density = UNFORMATED;
	  return 256;
	}
      else
	{
	  DPRINT("end ",0);
	  density = DOUBLE_DENS;
	  cmd.cmd = CMD_READ_ID | OPT_MFM_MODE | OPT_SET_HEAD;
	  work1 = sub1(dblBuf);
	  dblMaxNum = wrkMaxNum;
	  sglMaxNum = 0;
	  sub2();
	  return work1;
	}
    }
  else
    {
      DPRINT("end ",0);
      cmd.cmd = CMD_READ_ID | OPT_MFM_MODE;
      DPRINT("read ID MFM mode ",0);
      if (diskbios(&cmd) & return_ICL)
	{
	  DPRINT("end ",0);
	  density = SINGLE_DENS;
	  cmd.cmd = CMD_READ_ID | OPT_SET_HEAD;
	  work1 = sub1(sglBuf);
	  dblMaxNum = 0;
	  sglMaxNum = wrkMaxNum;
	  sub2();
	  return work1;
	}
      else
	{
	  DPRINT("end ",0);
	  density = MIXED_DENS;
	  cmd.cmd = CMD_READ_ID | OPT_SET_HEAD;
	  work1 = sub1(sglBuf);
	  sglMaxNum = wrkMaxNum;
	  cmd.cmd = CMD_READ_ID | OPT_MFM_MODE | OPT_SET_HEAD;
	  work2 = sub1(dblBuf);
	  dblMaxNum = wrkMaxNum;
	  work1 = (work1 > work2) ? work1 : work2;
	  sub2();
	  return work1;
	}
    }
}

static int sub1(idbuf *ptr)
{
  int work;

  cmd.duplicate = 0;
  cmd.buf_ptr = (uchar *)ptr;
  DPRINT("sub1 ",0);
  diskbios(&cmd);
  DPRINT("end ",0);
  work = 0;
  for (wrkMaxNum = 0; ptr[wrkMaxNum].counter; ++wrkMaxNum)
    if (work < ptr[wrkMaxNum].N)
      work = ptr[wrkMaxNum].N;
  return 0x80 << work;
}

static void sub2()
{
  int dblID, sglID, id;
  int i, j, last;

  dblID = sglID = 0;
  for (id = 0; id < sglMaxNum + dblMaxNum; ++id)
    {
      if (sglID == sglMaxNum)
	goto set_double;
      if (dblID == dblMaxNum)
	goto set_single;
      if (sglBuf[sglID].counter < dblBuf[dblID].counter)
	goto set_single;
    set_double:
      buf[id] = dblBuf[dblID++];
      buf[id].density = DOUBLE_DENS;
      continue;
    set_single:
      buf[id] = sglBuf[sglID++];
      buf[id].density = SINGLE_DENS;
      continue;
    }
  j = 0;
  for (i = id - 1; i > 0; --i)
    {
      if ((buf[i].C == buf[0].C) && (buf[i].H == buf[0].H) &&
	  (buf[i].R == buf[0].R) && (buf[i].N == buf[0].N) &&
	  (buf[i].density == buf[0].density))
	{
	  if (buf[i].counter <= timer_max)
	    break;
	  j = i;
	}
    }
  maxNum = j ? j : (i ? i : id );
  for (i = maxNum - 1; i >= 0; --i)
    {
      last = buf[i].duplicate;
      for (j = i - 1; j >= 0; --j)
	{
	  if ((buf[i].C == buf[j].C) && (buf[i].H == buf[j].H) &&
	      (buf[i].R == buf[j].R) && (buf[i].N == buf[j].N) &&
	      (buf[i].density == buf[j].density))
	    {
	      ++buf[i].duplicate;
	      buf[j].duplicate = 1;
	    }
	}
      if (last == 0 && buf[i].duplicate)
	++(buf[i].duplicate);
    }
  offset = 0;
}

int NonStdRead(uchar *ptr)
{
  if (density == UNFORMATED)
    {
      setmem((char *)ptr, 256, 0);
      return 256;
    }
  cmd.cmd = CMD_READ_DATA;
  if (buf[offset].density == DOUBLE_DENS)
    cmd.cmd |= OPT_MFM_MODE;
  if (buf[offset].duplicate)
    {
      cmd.cmd |= OPT_SET_HEAD;
      cmd.duplicate = buf[offset].duplicate - 1;
    }
  cmd.params[0] = cmd.params[4] = buf[offset].C;
  cmd.params[1] = buf[offset].H;
  cmd.params[2] = buf[offset].R;
  cmd.params[3] = buf[offset].N;
  cmd.params[5] = 1;
  cmd.params[6] = 0xFF;
  cmd.buf_len = 0x80 << buf[offset].N;
  cmd.buf_ptr = ptr;
  DPRINT("\nread data ",0);
  DPRINT("offset    %X ",offset);
  DPRINT("param C   %X ",cmd.params[0]);
  DPRINT("param H   %X ",cmd.params[1]);
  DPRINT("param R   %X ",cmd.params[2]);
  DPRINT("param N   %X ",cmd.params[3]);
  DPRINT("buf_len   %d ",cmd.buf_len);
  DPRINT("buf_ptr   %X ",cmd.buf_ptr);
  DPRINT("duplicate %X ",cmd.duplicate);
  diskbios(&cmd);
  DPRINT("end ",0);
  return cmd.buf_len;
}

int NonStdWrite(uchar *ptr)
{
  if (density == UNFORMATED)
    {
      setmem((char *)ptr, 256, 0);
      return 256;
    }
  cmd.cmd = CMD_WRITE_DATA;
  if (buf[offset].density == DOUBLE_DENS)
    cmd.cmd |= OPT_MFM_MODE;
  if (buf[offset].duplicate)
    {
      cmd.cmd |= OPT_SET_HEAD;
      cmd.duplicate = buf[offset].duplicate - 1;
    }
  cmd.params[0] = cmd.params[4] = buf[offset].C;
  cmd.params[1] = buf[offset].H;
  cmd.params[2] = buf[offset].R;
  cmd.params[3] = buf[offset].N;
  cmd.params[5] = 1;
  cmd.params[6] = 0xFF;
  cmd.buf_len = 0x80 << buf[offset].N;
  cmd.buf_ptr = ptr;
  DPRINT("write data ",0);
  diskbios(&cmd);
  DPRINT("end ",0);
  return cmd.buf_len;
}

int NonStdTrackSel()
{
drive_in:
  B_LOCATE(0, 24);
  B_CLR_ED();
  cmd.drive = 255;
  while (cmd.drive > 3)
    {
      B_LOCATE(36, 23);
      B_PRINT(MesDriveNoQ);
      B_LOCATE(59, 23);
      cmd.drive = GetNum(10, 1, 0);
      if (cmd.drive == 255)
	return 0;
    }
disktype_in:
  B_LOCATE(0, 25);
  B_CLR_ED();
  cmd.disktype = 2;
  while (cmd.disktype > 1)
    {
      B_LOCATE(31, 24);
      B_PRINT(MesDiskTypeQ);
      B_LOCATE(64, 24);
      switch(GetKey())
	{
	case 0x4f:/* 0 */
	case 0x0b:/* 0 */
	  cmd.disktype = 0;
	  B_PRINT("2HD");
	  break;
	case 0x4b:/* 1 */
	case 0x02:/* 1 */
	  cmd.disktype = 1;
	  B_PRINT("2DD/2D");
	  break;
	case 0x01:/* ESC */
	  goto drive_in;
	}
    }
/* track_in */
  cmd.track = 255;
  while (cmd.track > TRACK_MAX-1)
    {
      B_LOCATE(35, 25);
      B_PRINT(MesTrackNoQ);
      B_LOCATE(60, 25);
      cmd.track = GetNum(10, 3, 0);
      if (cmd.track == 255)
	goto disktype_in;
    }
  B_LOCATE(0, 23);
  B_CLR_ED();
  B_CUROFF();
  return 1;
}

void NonStdDriveIn()
{
  int c;

  c = 0x1000;
  while (c > 3)
    {
      B_LOCATE(35, 29);
      B_PRINT(MesDriveNoQ);
      B_LOCATE(58, 29);
      c = GetNum(10, 1, -1);
    }
  B_LOCATE(0, 29);
  B_ERA_AL();
  B_CUROFF();
  if (c != ERROR)
    cmd.drive = c;
}

void NonStdTrackIn()
{
  int c;

  c = 0x1000;
  while (c >= TRACK_MAX)
    {
      B_LOCATE(35, 29);
      B_PRINT(MesTrackNoQ);
      B_LOCATE(60, 29);
      c = GetNum(10, 3, -1);
    }
  B_LOCATE(0, 29);
  B_ERA_AL();
  B_CUROFF();
  if (c != ERROR)
    cmd.track = c;
}

void NonStdTrackHead()
{
  cmd.track = 0;
}

int NonStdTrackInc()
{
  if (++cmd.track == TRACK_MAX)
    {
      cmd.track = 0;
      return ERROR;
    }
  return 0;
}

int NonStdTrackDec()
{
  if (--cmd.track == 255)
    {
      cmd.track = TRACK_MAX-1;
      return ERROR;
    }
  return 0;
}

void NonStdIn()
{
  int c;

  c = 0x1000;
  while (c == 0 || c > maxNum)
    {
      B_LOCATE(35, 29);
      B_PRINT("\33[37msector number (1 - ");
      PrintDec(maxNum, 2);
      B_PRINT(") ?   ");
      B_LOCATE(60, 29);
      c = GetNum(10, 2, -1);
    }
  B_LOCATE(0, 29);
  B_ERA_AL();
  B_CUROFF();
  if (c != ERROR)
    offset = c - 1;
}

void NonStdHead()
{
  offset = 0;
}

int NonStdInc()
{
  if (++offset == maxNum)
    {
      offset = 0;
      return ERROR;
    }
  return 0;
}

int NonStdDec()
{
  if (--offset == -1)
    {
      offset = maxNum - 1;
      return ERROR;
    }
  return 0;
}

void NonStdPrint()
{
  B_LOCATE(8, 4);
  B_PRINT(MesDriveNo);
  B_COLOR(7);
  PrintDec(cmd.drive, 1);
  B_PRINT(Mes4Spaces);
  B_PRINT((cmd.disktype == 1) ? Mes2DD : Mes2HD);
  B_PRINT(Mes4Spaces);
  B_PRINT(MesTrackNo);
  B_COLOR(7);
  PrintDec(cmd.track, 3);
  B_PRINT(Mes4Spaces);
  B_PRINT("\33[35m");
  switch (density)
    {
    case SINGLE_DENS:
      B_PRINT(MesSingle);
      B_PRINT(Mes2Spaces);
      break;
    case DOUBLE_DENS:
      B_PRINT(MesDouble);
      B_PRINT(Mes2Spaces);
      break;
    case UNFORMATED:
      B_PRINT("unformated\33[33m");
      B_ERA_ED();
      B_LOCATE(0, 5);
      B_ERA_AL();
      B_CUROFF();
      return;
    default:
      B_PRINT((buf[offset].density == SINGLE_DENS) ? "mixed(s)" : "mixed(d)");
    }
  B_PRINT(Mes4Spaces);
  B_PRINT(MesResult);
  B_PUTC('$');
  PutHex(cmd.results[0] >> 4);
  PutHex(cmd.results[0]);
  B_PUTC(' ');
  PutHex(cmd.results[1] >> 4);
  PutHex(cmd.results[1]);
  B_PUTC(' ');
  PutHex(cmd.results[2] >> 4);
  PutHex(cmd.results[2]);
  B_LOCATE(82, 5);
  if (cmd.results[0] & 0x40)
    B_PRINT("\33[36mERRORS");
  else
    B_PRINT(Mes6Spaces);
  B_LOCATE(10, 5);
  B_PRINT("\33[31mNo.        b     g     q     m     timer      -      (    )\33[37m");
  B_LOCATE(14, 5);
  PrintDec(offset + 1, 2);
  if (buf[offset].duplicate)
    {
      B_LOCATE(16, 5);
      B_PUTC('(');
      PrintDec(buf[offset].duplicate, 2);
      B_PUTC(')');
    }
  B_LOCATE(24, 5);
  PrintDec(buf[offset].C, 3);
  B_LOCATE(31, 5);
  PrintDec(buf[offset].H, 3);
  B_LOCATE(38, 5);
  PrintDec(buf[offset].R, 3);
  B_LOCATE(45, 5);
  PrintDec(buf[offset].N, 3);
  B_LOCATE(55, 5);
  PrintDec(buf[offset].counter, 4);
  B_LOCATE(62, 5);
  PrintDec(buf[offset + 1].counter, 4);
  B_LOCATE(68, 5);
  PrintDec(buf[offset + 1].counter - buf[offset].counter, 4);
  B_COLOR(3);
  B_CUROFF();
}
